//Copyright (c) 2006-2007 Microsoft Corporation.  All rights reserved
using System;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Web;

using InvocationFramework.Common;
using InvocationFramework.Common.Instrumentation;
using InvocationFramework.Common.Exceptions;
using InvocationFramework.Common.Library;
using InvocationFramework.Common.Resources;

namespace InvocationFramework.Adapters.Inproc
{
    /// <summary>
    /// Esta clase maneja todos los aspectos del la funcionalidad del servicio de "Hot Deploy"
    /// </summary>
    /// <remarks>
    /// Los siguientes metodos privados conforman el "Servicio de Hot Deploy" antes mencionado".
    /// Basicamente funciona de la siguiente manera:
    /// Cuando el InprocAdapter es inicializado, este verfica si el atributo de configuracion "enableHotDeploy" se encuentra habilitado.
    /// En caso de ser as, la propiedad CatalogManager (responsable por llevar la lista de servicios ejecutables y ejecutar a los mismos)
    /// es cargada en un AppDomain diferente.
    /// Luego, se activa un FileSystemWatcher que monitorea los eventos de la carpeta donde estn los assemblies de servicios
    /// En caso de que algn assebly sea sobrescrito, eliminado o agregado en la carpeta, autmticamente se dispara un evento
    /// que descarga el AppDomain, eliminado al CatalogManager para luego reincializarlo con los nuevos valores.
    /// </remarks>
    /// <author>Rodolfo Finochietti - Sergio Borromei 16/05/2005 13.18</author>
    internal class HotDeployManager
    {               
        #region Constants

        /// <summary>
        /// Reintentos por si falla una operacion de IO
        /// </summary>
        private const int Retries = 50;

        #endregion

        #region Constructors

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="servicesFolder">Carpeta donde se encuentran los servicios</param>
        /// <param name="workingFolder">Carpeta temporal de trabajo</param>
        /// <param name="enableHotDeploy">Indica si esta habilitado el hot deploy</param>
        public HotDeployManager( string servicesFolder, string workingFolder, bool enableHotDeploy )
        {
            //Seteo las carpetas de trabajo para el servicio de Hot Deploy (difieren de las anteriores si estoy en un contexto web)
            //y las valido. Si no existen se arroja una excepcin.
            //Esto era diferente hasta la version 20060228, en la que no se validaban las carpetas si el hot deploy estaba deshabilitado.
            SetUpHotDeployFolders( servicesFolder, workingFolder, Environment.CurrentDirectory );

            //sborrome@microsoft.com 
            //Si el Hot Deploy est habilitado, entonces creo el catlogo de servicios en otro App Domain (consume mas recursos),
            //y levanto el FileSystemWatcher correspondiente para que lo re-inicialice en caso de una actualizacin en hot.
            if ( enableHotDeploy )
            {
                InitHotDeployEnvironment( );
            }
            else
            {
                _catalogManager = new CatalogManager( );
                _catalogManager.PopulateCatalog( _hotDeployServicesFolder, _hotDeployWorkingFolder );
            }

            //Inicio el resolver de referencia de assemblies, para los casos en los que hay dependencia que estan en el mismo directorio de servicios
            AssemblyResolver.Start( _hotDeployWorkingFolder );
        }

        #endregion

        #region Private Vars

        /// <summary>
        /// AppDomain en donde se almacenar el catlogo de servicios en caso de que el Hot Deploy se encuentre activado.
        /// El catlogo de servicios contiene la lista de servicios y operaciones que este adaptador puede ejecutar.
        /// Adicionalmente provee la funcionalidad de ejecutar operaciones bajo contexto transaccional o no.
        /// En caso de usar Hot Deploy, cuando se sobrescribe una .dll con servicios, automaticamente se baja este dominio 
        /// y se reinicializa el catalogo.
        /// </summary>
        private AppDomain _hotDeployDomain;

        /// <summary>
        /// Esta variable se usa para que el monitor de hot deploy sepa cual es el directorio de servicios
        /// </summary>
        private string _hotDeployServicesFolder;

        /// <summary>
        /// Esta variable se usa para que el monitor de hot deploy sepa cual es el directorio de trabajo
        /// </summary>
        private string _hotDeployWorkingFolder;

        /// <summary>
        /// Esta variable se usa para que el monitor de hot deploy sepa cual es el donde se ejecuta la aplicacion
        /// </summary>
        private string _hotDeployApplicacionBase;

        /// <summary>
        /// Monitor para el directorio donde se encuentran los servicios. Solamente se usa si el Hot Deploy est activado.
        /// </summary>
        private FileSystemWatcher _servicesFolderMonitor;

        
        #endregion

        #region Properties

        /// <summary>
        /// Variable privada para el manejo de la propiedad
        /// </summary>
        private CatalogManager _catalogManager;

        /// <summary>
        /// El administrador de catlogo de servicios contiene la lista de servicios y operaciones que este adaptador puede ejecutar.
        /// Adicionalmente provee la funcionalidad de ejecutar las operaciones de los servicios, ya sea bajo contexto transaccional o no.
        /// En caso de usar Hot Deploy, el CatalogManager es creado en otro AppDomain, a efectos de facilitar la actualizacin 
        /// de servicios sin detener al sistema.
        /// </summary>
        public CatalogManager Catalog
        {
            get
            {
                return _catalogManager;
            }
        }

        #endregion

        #region Private Methods

        /// <summary>
        /// Metodo privado para setear las carpetas de trabajo del servicio de Hot Deploy.
        /// Es llamado solamente desde el Init()
        /// </summary>
        /// <param name="servicesFolder"></param>
        /// <param name="workingFolder"></param>
        /// <param name="applicacionBase"></param>
        private void SetUpHotDeployFolders( string servicesFolder, string workingFolder, string applicacionBase )
        {
            //Si estoy en un contexto http los paths pueden relativos al directorio virtual del sitio y el path de ejecucion del nuevo app domain es '\Bin'
            if ( HttpContext.Current == null )
            {
                _hotDeployServicesFolder = servicesFolder;
                _hotDeployWorkingFolder = workingFolder;
                _hotDeployApplicacionBase = applicacionBase;
            }
            else
            {
                //Verifico si el path ingresado es relativo o absoluto
                if ( Directory.Exists( servicesFolder ) )
                    _hotDeployServicesFolder = servicesFolder;
                else
                    _hotDeployServicesFolder = HttpContext.Current.Server.MapPath( servicesFolder );

                //Verifico si el path ingresado es relativo o absoluto
                if ( Directory.Exists( workingFolder ) )
                    _hotDeployWorkingFolder = workingFolder;
                else
                    _hotDeployWorkingFolder = HttpContext.Current.Server.MapPath( workingFolder );

                //El directorio de la aplicacion es el bin del sitio
                _hotDeployApplicacionBase = HttpContext.Current.Server.MapPath( "Bin" );
            }

            //Valido que los directorios existan
            if ( !Directory.Exists( _hotDeployServicesFolder ) )
                throw new InvocationFrameworkException( String.Format( CultureInfo.InvariantCulture, Texts.CatalogManagerServiceDirectoryNotExist, _hotDeployServicesFolder ) );
            if ( !Directory.Exists( _hotDeployWorkingFolder ) )
                throw new InvocationFrameworkException( String.Format( CultureInfo.InvariantCulture, Texts.CatalogManagerWorkingDirectoryNotExist, _hotDeployWorkingFolder ) );
            if ( _hotDeployServicesFolder.Equals( _hotDeployWorkingFolder, StringComparison.InvariantCultureIgnoreCase ) )
                throw new InvocationFrameworkException( String.Format( CultureInfo.InvariantCulture, Texts.CatalogManagerWorkingServicesDirectorysAreEquals, _hotDeployServicesFolder, _hotDeployWorkingFolder ) );
        }

        private void InitHotDeployEnvironment( )
        {
            Guid instance = Guid.NewGuid( );

            TraceHelper.Verbose( Texts.HotDeployManagerCreateNewDomainStartTrace, instance );

            //Creo un nuevo application domain para correr los servicio y soportar el hot deploy
            _hotDeployDomain = ReflectionHelper.CreateNewDomain( String.Format( CultureInfo.InvariantCulture, "InvocationFrameworkInprocServices:{0}", instance ), "InvocationFrameworkInprocServicesDomain", _hotDeployApplicacionBase, AppDomain.CurrentDomain.SetupInformation.ConfigurationFile );

            //Creo una instancia del catalogo que corra en el nuevo application domain
            _catalogManager = ( CatalogManager )ReflectionHelper.CreateObjectInAnotherDomain( _hotDeployDomain, "InvocationFramework.Adapters.Inproc", "InvocationFramework.Adapters.Inproc.CatalogManager" );

            //Lleno el catalogo
            _catalogManager.PopulateCatalog( _hotDeployServicesFolder, _hotDeployWorkingFolder );

            //Activo el monitor de hot deploy
            StartHotDeployMonitor( );
        }

        /// <summary>
        /// Inicia el monitoreo para hot deploy de servicios
        /// </summary>
        private void StartHotDeployMonitor( )
        {
            if ( _servicesFolderMonitor == null || !_servicesFolderMonitor.EnableRaisingEvents )
            {
                //Configuro el watcher
                _servicesFolderMonitor = new FileSystemWatcher( );
                _servicesFolderMonitor.Path = _hotDeployServicesFolder;
                _servicesFolderMonitor.NotifyFilter = NotifyFilters.LastAccess | NotifyFilters.LastWrite | NotifyFilters.FileName | NotifyFilters.Size;
                _servicesFolderMonitor.Filter = Constants.ServicesFileFilter;

                //Me susbscribo a los eventos
                _servicesFolderMonitor.Created += new FileSystemEventHandler( this._servicesFolderMonitor_Created );
                _servicesFolderMonitor.Changed += new FileSystemEventHandler( this._servicesFolderMonitor_Changed_Deleted );
                _servicesFolderMonitor.Deleted += new FileSystemEventHandler( this._servicesFolderMonitor_Changed_Deleted );

                //Empiezo a monitorear
                _servicesFolderMonitor.EnableRaisingEvents = true;

                TraceHelper.Verbose( Texts.HotDeployManagerStartTrace );
            }
        }

        /// <summary>
        /// Este evento se dispara cuando se agrega un assembly al directorio de servicios
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage( "Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes" )]
        private void _servicesFolderMonitor_Created( object sender, FileSystemEventArgs e )
        {
            string file = e.FullPath;
            string tempFile = Path.Combine( _hotDeployWorkingFolder, Path.GetFileName( file ) );

            //Reintento por que muchas veces el file system watcher avisa cuando todavia el archivo
            //no termino de copiarse
            for ( int i = 0; i < Retries; i++ )
            {
                try
                {
                    //Copio el archivo al repositorio directorio temporal el ejecucion de assemblys
                    File.Copy( file, tempFile );

                    //Si llego hasta aca la copia fue exitosas
                    break;
                } 
                catch { }
            }

            _catalogManager.AddAssemblyToCatalog( tempFile );

            TraceHelper.Verbose( String.Format( CultureInfo.InvariantCulture, Texts.HotDeployManagerAddTrace, tempFile ) );
        }

        /// <summary>
        /// Este evento se dispara cuando se elimina un assembly en el directorio de servicios
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void _servicesFolderMonitor_Changed_Deleted( object sender, FileSystemEventArgs e )
        {
            AppDomain.Unload( _hotDeployDomain );
            InitHotDeployEnvironment( );

            TraceHelper.Verbose( String.Format( CultureInfo.InvariantCulture, Texts.HotDeployManagerUnloadAppDomain, e.FullPath ) );
        }

        #endregion
    }
}