//Copyright (c) 2006-2007 Microsoft Corporation.  All rights reserved
using System;
using System.Globalization;
using System.Xml;

using InvocationFramework.Common;
using InvocationFramework.Common.Exceptions;
using InvocationFramework.Common.Instrumentation;
using InvocationFramework.Common.Interfaces;
using InvocationFramework.Common.Resources;

namespace InvocationFramework.Adapters.Inproc
{
    /// <summary>
    /// Adaptador para ejecutar servicios localmente en el mismo proceso del Invocation Framework
    /// </summary>
    /// <remarks>
    /// Los servicios que ejecuta este adaptador se encuentra implementados en clases .Net marcadas con los atributos que se encuentran
    /// en el namespace InvocationFramework.Adapters.Inproc.Attributes
    /// </remarks>
    /// <author>Rodolfo Finochietti 01/11/2005 13.18</author>
    public class InprocAdapter : IServiceAdapter
    {
        #region Private Vars

        /// <summary>
        /// Instancia del manager del servicio de hot deploy
        /// </summary>
        private HotDeployManager _hotDeployManager;

        #endregion

        #region Constructors

        /// <summary>
        /// Constructor por defecto
        /// </summary>
        public InprocAdapter( ) { }

        #endregion

        #region IServiceAdapter Members

        /// <summary>
        /// Variable privada para el manejo de la propiedad
        /// </summary>
        private string _id;

        /// <summary>
        /// Identificador del adaptador
        /// </summary>
        public string Id
        {
            get
            {
                return _id;
            }
        }

        /// <summary>
        /// Variable privada para el manejo de la propiedad
        /// </summary>
        private bool _initialized;

        /// <summary>
        /// Indica si el adaptador esta inicializado o no
        /// </summary>
        public bool Initialized
        {
            get
            {
                return _initialized;
            }
        }

        /// <summary>
        /// Inicializa el adaptador
        /// </summary>
        /// <param name="id">Identificador del adaptador</param>
        /// <param name="configuration">Configuracion del adaptador</param>
        public void Init( string id, XmlNode configuration )
        {
            if ( !_initialized )
            {
                try
                {
                    //Controlo que los parametros no sean nulos
                    if ( String.IsNullOrEmpty( id ) )
                        throw new ArgumentNullException( "id", String.Format( CultureInfo.InvariantCulture, Texts.InprocAdapterArgumentNull, "id" ) );

                    if ( configuration == null )
                        throw new ArgumentNullException( "configuration", String.Format( CultureInfo.InvariantCulture, Texts.InprocAdapterArgumentNull, "configuration" ) );

                    //Cargo las propiedades
                    _id = id;

                    //Cargo el catalogo con los servicios
                    string servicesFolder = configuration.Attributes[ "servicesFolder" ].Value;
                    string workingFolder = configuration.Attributes[ "workingFolder" ].Value;

                    //El atributo enableHotDeploy es opcional
                    bool hotDeployEnabled = ( configuration.Attributes[ "enableHotDeploy" ] == null ? false : Convert.ToBoolean( configuration.Attributes[ "enableHotDeploy" ].Value, CultureInfo.InvariantCulture ) );

                    //Inicalizo el servicio de hot deploy
                    _hotDeployManager = new HotDeployManager( servicesFolder, workingFolder, hotDeployEnabled );

                    //Marco el flag de inicializado
                    _initialized = true;

                    TraceHelper.Verbose( String.Format( CultureInfo.InvariantCulture, Texts.InprocAdapterInitTrace, servicesFolder, workingFolder, hotDeployEnabled ) );
                }
                catch ( Exception ex )
                {
                    TraceHelper.Error( ex, Texts.InprocAdapterInitError );
                    throw;
                }
            }
        }

        /// <summary>
        /// Permite saber si un servicio esta disponible para ser ejecutado mediante este adaptador
        /// </summary>
        /// <param name="serviceName">Nombre del servicio</param>
        /// <returns>True este adaptador puede ejecutar este servicio, False de lo contrario</returns>
        public bool IsServiceAvailable( string serviceName )
        {
            if ( !_initialized )
                throw new InvocationFrameworkException( Texts.InprocAdapterNotInitialized );

            //Busco el servicio utilizando el catalogo
            return _hotDeployManager.Catalog.ContainsService( serviceName );
        }

        /// <summary>
        /// Verifica si existe una operacion en un servicio
        /// </summary>
        /// <param name="serviceName">Nombre del servicio</param>
        /// <param name="operationName">Nombre de la operacion</param>
        /// <returns>True si la operacion existe en el servicio, de lo contrario False</returns>
        public bool IsServiceOperationAvailable( string serviceName, string operationName )
        {
            if ( !_initialized )
                throw new InvocationFrameworkException( Texts.InprocAdapterNotInitialized );

            //Busco la operacion utilizando el catalogo
            return _hotDeployManager.Catalog.ContainsServiceOperation( serviceName, operationName );
        }

        /// <summary>
        /// Devuelve una lista con los nombres de los servicios mas utilizados
        /// </summary>
        /// <returns>Una lista con los nombres de los servicios mas utilizados</returns>
        public string[] GetMostRequestedServices( )
        {
            if ( !_initialized )
                throw new InvocationFrameworkException( Texts.InprocAdapterNotInitialized );

            //Le pregunto al catalogo cuales son los servicios muy utilizados
            return _hotDeployManager.Catalog.MostRequestedServices;
        }

        /// <summary>
        /// Ejecuta la operacion que se encuentra en un determinado servicio
        /// </summary>
        /// <param name="serviceName">Nombre del servicio</param>
        /// <param name="operationName">Nombre de la operacion a ejecutar</param>
        /// <param name="context">El contexto que se desea propagar junto con al ejecucion del servicios</param>
        /// <param name="parameters">Parametros de la operacion</param>
        /// <returns>Resultado de la ejecucion de la operacion</returns>
        public object Execute( string serviceName, string operationName, ContextDictionary context, params object[] parameters )
        {
            if ( !_initialized )
                throw new InvocationFrameworkException( Texts.InprocAdapterNotInitialized );

            //Controlo que exista el servicio en el catalog
            if ( !_hotDeployManager.Catalog.ContainsService( serviceName ) )
                throw new InvocationFrameworkException( String.Format( CultureInfo.InvariantCulture, Texts.InprocAdapterServiceNotFound, serviceName ) );

            //Constrolo que el servicio tenga la operacion que se quiere ejecutar
            if ( !_hotDeployManager.Catalog.ContainsServiceOperation( serviceName, operationName ) )
                throw new InvocationFrameworkException( String.Format( CultureInfo.InvariantCulture, Texts.InprocAdapterOperationNotFound, serviceName, operationName ) );

            //La ejecucion se hace en la clase CatalogManager. Si el servicio de Hot Deploy se encuentra habilitado, 
            //la propiedad _hotDeployManager.Catalog es creada en otro AppDomain durante la inicilizacion.
            //De esta forma, al actualizar el catlogo, el servicio de Hot Deploy reinicializa el AppDomain, dando
            //una finalizacin prolija a las ejecuciones en curso.
            try
            {
                return _hotDeployManager.Catalog.Execute( serviceName, operationName, parameters );
            }
            catch ( Exception ex )
            {
                //Si se produjo un excepcion, puede ser que la excepcion real este dentro de la inner por que la ejecucion se hace por reflection
                if ( ex.InnerException == null )
                    throw;
                else
                    throw ex.InnerException;
            }
        }

        #endregion
    }
}