//Copyright (c) 2006-2007 Microsoft Corporation.  All rights reserved
using System;
using System.Xml;

using InvocationFramework.Common;

namespace InvocationFramework.Common.Interfaces
{
    /// <summary>
    /// Represents an adapter to interact with a service provider.
    /// </summary>
    /// <remarks>
    /// This is the main interface of SIF and provides the contract definition to extend the architecture to new service providers.
    /// To create and use new service adapter you just have to implement this interface and declare your implementation inside SIF configuration section.
    /// When the first service is invoked, SIF reads the configuration and creates an instance of each adapter configured in the section serviceAdapters.
    /// For each adapter instance, SIF calls to the Init() method passing an XmlNode with the configuration sub-section of the adapter; in other words,
    /// the inner xml of the serviceAdapter element. 
    /// After initializing, each adapter instance is stored in the Service Adapter Cache for later service invocations.
    /// SIF also has an internal Service Catalog, which consists of a list containing all the services that can be invoked 
    /// and the ID of the corresponding Service Adapter. After adapter initialization, SIF calls to GetMostRequestedServices() method
    /// to populate the Service Catalog.
    /// During service invokation process, SIF queries the Service Catalg to determine which Service Adapter can be used to invoke the requested service.
    /// If service adapter cant be detemined (because no entries were found for that service), SIF 'asks' each Service Adapter, if the requested service
    /// is available by calling IsServiceAvailable() or IsServiceOperationAvailable() methods.
    /// </remarks>
    /// <example>
    ///     This is a sample configuration section.
    ///     <servicesInvocationFramework>
    ///         <caller>
    ///            <serviceAdapters>
    ///                <serviceAdapter id="WebServicesAdapter" order="0" enabled="true" type="InvocationFramework.Adapters.WebServices.WebServicesAdapter, InvocationFramework.Adapters.WebServices, Version=2.0.0.0, Culture=neutral, PublicKeyToken=4396b271bcbb0166">
    ///                    <services>
    ///                        <service name="BasicService" wsdlLocation="http://localhost:2597/TestServices/BasicService.asmx?WSDL" contractType="Contracts.IBasicService, Contracts"/>
    ///                        <service name="DataService" wsdlLocation="http://localhost:2597/TestServices/DataService.asmx?WSDL" contractType="Contracts.IDataService, Contracts"/>
    ///                        <service name="DTOService" wsdlLocation="http://localhost:2597/TestServices/DTOService.asmx?WSDL" contractType="Contracts.IDTOService, Contracts"/>
    ///                    </services>
    ///                </serviceAdapter>
    ///            </serviceAdapters>
    ///        </caller>
    ///     </servicesInvocationFramework>
    /// </example>
    public interface IServiceAdapter
    {
        /// <summary>
        /// Adapter ID. Contains the id attribute defined in serviceAdapter configuration section.
        /// </summary>
        /// <remarks>
        /// This ID must be unique among the adapters configured in the serviceAdapters section.
        /// </remarks>
        string Id
        {
            get;
        }

        /// <summary>
        /// Inidcates if the Service Adapter is ready to be used.
        /// Must be set to true by Service Adapter after initialization.
        /// </summary>
        bool Initialized
        {
            get;
        }

        /// <summary>
        /// Inits the adapter. This method is called by SIF when the first service invocation occurs.
        /// </summary>
        /// <param name="id">Identificador del adaptador</param>
        /// <param name="configuration">Configuracion del adaptador</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage( "Microsoft.Design", "CA1059:MembersShouldNotExposeCertainConcreteTypes" )]
        void Init( string id, XmlNode configuration );

        /// <summary>
        /// Allows SIF to know if a particular service can be invoked by this adapter.
        /// </summary>
        /// <remarks>
        /// This method is used by SIF to ask the adapter if a particular service can be invoked.
        /// </remarks>
        /// <param name="serviceName">Name of the service</param>
        /// <returns>Must return true if service can be invoked by this adapter. Otherwise, false.</returns>
        bool IsServiceAvailable( string serviceName );

        /// <summary>
        /// Allows SIF to know if a particular service operation can be invoked by this adapter
        /// </summary>
        /// <param name="serviceName">Name of the service</param>
        /// <param name="operationName">Name of the operation</param>
        /// <returns>Must return true if service operation can be invoked by this adapter. Otherwise, false.</returns>
        bool IsServiceOperationAvailable( string serviceName, string operationName );

        /// <summary>
        /// Returns a list of the services more likely to be invoked.
        /// </summary>
        /// <remarks>
        /// This method is used by SIF during initialization to populate the Service Catalog. 
        /// </remarks>
        /// <returns></returns>
        string[] GetMostRequestedServices( );

        /// <summary>
        /// Invokes a service operation.
        /// </summary>
        /// <param name="serviceName">Name of the service</param>
        /// <param name="operationName">Name of the operation</param>
        /// <param name="context">Context data, used to send out-of-band parameters.</param>
        /// <param name="parameters">Payload required by the operation</param>
        /// <returns>Execution results</returns>
        object Execute( string serviceName, string operationName, ContextDictionary context, params object[] parameters );
    }
}