//Copyright (c) 2006 Microsoft Corporation.  All rights reserved
using System;
using System.Configuration;
using System.Collections.Generic;
using System.Globalization;
using System.Runtime.CompilerServices;
using System.Xml;

using InvocationFramework.Common;
using InvocationFramework.Common.Exceptions;
using InvocationFramework.Common.Instrumentation;
using InvocationFramework.Common.Interfaces;
using InvocationFramework.Common.Library;
using InvocationFramework.Common.Resources;

using InvocationFramework.Core.Configuration;

namespace InvocationFramework.Core
{
    /// <summary>
    /// Virtual Proxy to configured service adapters.
    /// </summary>
    /// <remarks>
    /// This class represents the implementation of the Virtual Proxy pattern (http://www.dofactory.com/Patterns/PatternProxy.aspx) 
    /// where the "subject" is the IServiceAdapter interface and the "real-subjects" are the configured service adapters.
    /// This is an internal class, and is intended to be used only by the ServiceCaller facade to access the service adapter sub-system.
    /// </remarks>
    /// <author>Rodolfo Finochietti 24/08/2005 9.43</author>
    internal sealed class ServiceBridge : IServiceAdapter
    {
        #region Private Vars

        /// <summary>
        /// Adapter cach.
        /// </summary>
        /// <remarks>
        /// During initialization, the ServiceBridge creates an instance of each configured service adapter. 
        /// Those instances are stored in memory through the life cycle of the ServiceBridge in order to 
        /// dispatch the service invocations.
        /// </remarks>
        private AdapterCache _adapterCache;

        #endregion

        #region IServiceAdapter Members

        private string _id;

        /// <summary>
        /// Adapter ID. 
        /// </summary>
        /// <remarks>
        /// This property is initializaded by ServiceCaller during initialization with the value #ServiceCaller#.
        /// </remarks>
        public string Id
        {
            get
            {
                return _id;
            }
        }

        /// <summary>
        /// Private var to store property state.
        /// </summary>
        private bool _initialized;

        /// <summary>
        /// States if adapter was succesfully initializated.
        /// </summary>
        public bool Initialized
        {
            get
            {
                return _initialized;
            }
        }

        /// <summary>
        /// Inits the ServiceBridge loading the configured adapters and enabling the performance counters.
        /// </summary>
        /// <param name="id">ID of the adapter. In this case, the ServiceCaller facade calls this method sending "#ServiceCaller#" as the id.</param>
        /// <param name="configuration">Xml configuration node containing metadata of configured service adapters</param>
        /// <remarks>
        /// This method is part of the implementation of IServiceAdapter interface and is called by thge ServiceCaller facade. 
        /// The first task of this method is to enable all the performance counters provided by SIF. The AdapterCache instance is then created, 
        /// and populated with the service adapter instances.
        /// Adapters instances are created and initialized one by one for each adapter element in the configuration.
        /// After successful initialization (no exceptions were raised), the Initialized property is set to true.
        /// </remarks>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        public void Init(string id, XmlNode configuration)
        {
            if (!_initialized)
            {
                try
                {
                    if (String.IsNullOrEmpty(id))
                        throw new ArgumentNullException("id", String.Format(CultureInfo.InvariantCulture, Texts.ServiceBridgeArgumentNull, "id"));

                    if (configuration == null)
                        throw new ArgumentNullException("configuration", String.Format(CultureInfo.InvariantCulture, Texts.ServiceBridgeArgumentNull, "configuration"));

                    //Levanto la configuracion
                    ConfigurationHelper config = new ConfigurationHelper();

                    //Configura la instrumentacion
                    PerformanceCountersManager.Instance.EnablePerformanceCounters = config.EnablePerformanceCounters;

                    //Cargo las propiedades
                    _id = id;

                    //Inicializo el cache de adaptadores
                    _adapterCache = new AdapterCache(config.EnableLocalCatalog, config.LocalCatalogInvalidationInterval);

                    //Cargo los adaptadores que correspondan segun la configuracion pasada (o los del cliente o lo de servidor)
                    List<AdapterElement> adapters = null;
                    if (Convert.ToBoolean(configuration.FirstChild.Attributes["publisherAdapters"].Value, CultureInfo.InvariantCulture))
                    {
                        adapters = config.PublisherAdapters;
                        TraceHelper.Verbose(Texts.ServiceBridgePublisherAdapters);
                    }
                    else
                    {
                        adapters = config.CallerAdapters;
                        TraceHelper.Verbose(Texts.ServiceBridgeCallerAdapters);
                    }

                    if (adapters != null)
                    {
                        //Cargo los adaptadores segun corresponda
                        foreach (AdapterElement adapter in adapters)
                        {
                            //Si el adapter esta habilitado en la configuracion lo agrego al cache de adapters
                            if (adapter.Enabled)
                            {
                                //Creo el adapter con el type especificado en la configuracion
                                string adapterType = adapter.Type;
                                IServiceAdapter adapterInstance = null;

                                try
                                {
                                    adapterInstance = ReflectionHelper.CreateObject(adapterType) as IServiceAdapter;
                                }
                                catch (Exception ex)
                                {
                                    TraceHelper.Error(ex, String.Format(CultureInfo.InvariantCulture, Texts.ServiceBridgeAdapterNotLoad, adapterType));
                                }

                                if (adapterInstance != null)
                                //Si pude levantar el adapter lo agrego a la coleccion de adapters
                                {
                                    //Inicializo el adapter
                                    adapterInstance.Init(adapter.Id, adapter.AdapterConfiguration);

                                    //Agrego el adapter al cache de adapters
                                    _adapterCache.AddAdapter(adapterInstance, adapter.Order);
                                }
                            }
                        }

                        //Actualizo el catalogo con los servicios mas utilizados de todos los adaptadores cargados previamente
                        _adapterCache.UpdateCatalogWithMostRequestedServices();
                    }

                    //Libero recursos
                    config = null;

                    _initialized = true;
                }
                catch (Exception ex)
                {
                    TraceHelper.Error(ex, Texts.ServiceBridgeInitError);
                    throw;
                }
            }
        }

        /// <summary>
        /// Look up the service catalog to find the service name passed by parameter, returning true if found.
        /// </summary>
        /// <param name="serviceName">Name of the service to find</param>
        /// <returns>True if exists a service adapter in catalog which provides the requested service.</returns>
        public bool IsServiceAvailable(string serviceName)
        {
            if (!_initialized)
                throw new InvocationFrameworkException(Texts.ServiceBridgeNotInitialized);

            //Si no hay ningun adapter para atender a este servicio, el servicio no esta disponible
            if (_adapterCache.GetAdapterByService(serviceName) == null)
                return false;
            else
                return true;
        }

        /// <summary>
        /// Look up the service catalog to find the service operation passed by parameter, returning true if found.
        /// </summary>
        /// <param name="serviceName">Name of the service to find</param>
        /// <param name="operationName">Name of the operation to find</param>
        /// <returns>True if exists a service adapter in catalog which provides the requested service.</returns>
        public bool IsServiceOperationAvailable(string serviceName, string operationName)
        {
            if (!_initialized)
                throw new InvocationFrameworkException(Texts.ServiceBridgeNotInitialized);

            //Si no hay ningun adapter para atender a este servicio, el servicio no esta disponible
            if (_adapterCache.GetAdapterByServiceOperation(serviceName, operationName) == null)
                return false;
            else
                return true;
        }

        /// <summary>
        /// Returns a list with the names of "most requested services". 
        /// </summary>
        /// <returns>String array with the names of the services.</returns>
        /// <remarks>
        /// This method returns the service names stored in memory service catalog. In fact, returns a string array,
        /// result of calling the GetMostRequestedServices() method of each service adapter.
        /// </remarks>
        public string[] GetMostRequestedServices()
        {
            if (!_initialized)
                throw new InvocationFrameworkException(Texts.ServiceBridgeNotInitialized);

            return _adapterCache.GetCatalog();
        }

        /// <summary>
        /// Invokes a service operation calling the service adapter that provides access to the service.
        /// </summary>
        /// <param name="serviceName">Name of the service which holds the operation.</param>
        /// <param name="operationName">Name of the service operation to invoke.</param>
        /// <param name="context">Context data to pass out-of-band parameters to service provider.</param>
        /// <param name="parameters">Payload for the operation</param>
        /// <returns>Object with execution results.</returns>
        /// <remarks>
        /// 
        /// </remarks>
        public object Execute(string serviceName, string operationName, ContextDictionary context, params object[] parameters)
        {
            if (!_initialized)
                throw new InvocationFrameworkException(Texts.ServiceBridgeNotInitialized);

            //Busco el adapter que ejecuta el servicio en el cache
            string adapterId = _adapterCache.GetAdapterByService(serviceName);

            //Si no encuentro es un error
            if (adapterId == null)
                throw new InvocationFrameworkException(String.Format(CultureInfo.InvariantCulture, Texts.ServiceBridgeServiceNotFound, serviceName));

            //Obtengo el adapter para ejecutar la operacion
            IServiceAdapter adapter = _adapterCache.GetAdapter(adapterId);

            //Mido el inicio de la ejecucion
            PerformanceCountersManager.StartStopWatch();

            //Ejecuto la operacion
            object result = adapter.Execute(serviceName, operationName, context, parameters);

            //Mido el tiempo final e incremento los performance counters
            PerformanceCountersManager.Instance.IncrementInvocationExecutionCounterByOne(String.Format(CultureInfo.InvariantCulture, "{0}\\{1}", serviceName, operationName), PerformanceCountersManager.StopStopWatch());

            //Si tengo un adapter mando a ejecutar el servicio
            return result;
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Allows service invocation avoiding searches in service catalog.
        /// </summary>
        /// <param name="adapterId">ID of the adapter which provides access to service</param>
        /// <param name="serviceName">Name of the service that holds the operation</param>
        /// <param name="operationName">Name of the service operation to invoke</param>
        /// <param name="context">Context data to pass out-of-band parameters to service provider</param>
        /// <param name="parameters">Payload for the operation</param>
        /// <returns>Object with execution results.</returns>
        public object FixAdapterExecute(string adapterId, string serviceName, string operationName, ContextDictionary context, params object[] parameters)
        {
            if (!_initialized)
                throw new InvocationFrameworkException(Texts.ServiceBridgeNotInitialized);

            //Controlo que el parametro no sea nulo
            if (String.IsNullOrEmpty(adapterId))
                throw new ArgumentNullException("adapterId", String.Format(CultureInfo.InvariantCulture, Texts.ServiceBridgeArgumentNull, "adapterId"));

            //Obtengo el adapter para ejecutar la operacion
            IServiceAdapter adapter = _adapterCache.GetAdapter(adapterId);

            //Controlo que se halla podido obtener el adaptador
            if (adapter == null)
                throw new InvocationFrameworkException(String.Format(CultureInfo.InvariantCulture, Texts.ServiceBridgeInvalidAdapter, adapterId));

            //Mido el inicio de la ejecucion
            PerformanceCountersManager.StartStopWatch();

            //Ejecuto la operacion
            object result = adapter.Execute(serviceName, operationName, context, parameters);

            //Mido el tiempo final e incremento los performance counters
            PerformanceCountersManager.Instance.IncrementInvocationExecutionCounterByOne(String.Format(CultureInfo.InvariantCulture, "{0}\\{1}", serviceName, operationName), PerformanceCountersManager.StopStopWatch());

            //Si tengo un adapter mando a ejecutar el servicio
            return result;
        }

        #endregion
    }
}