//Copyright (c) 2006-2007 Microsoft Corporation.  All rights reserved
using System;
using System.Text;
using System.Collections.Generic;
using System.Runtime.Remoting;
using System.Diagnostics;
using System.Threading;
using System.Xml;

using Microsoft.VisualStudio.TestTools.UnitTesting;

using InvocationFramework.Core.ProxyFactory;
using InvocationFramework.Common.Interfaces;
using InvocationFramework.Common.Exceptions;
using InvocationFramework.Adapters.Remoting;

using RemotingAdapterTests.Contracts;

namespace RemotingAdapterTest.UnitTests
{
    [TestClass]
    public class RemotingAdapterTests
    {
        public RemotingAdapterTests( ) { }

        //Remoting host process
        private static Process _server;

        [ClassInitialize( )]
        public static void MyClassInitialize( TestContext testContext )
        {
            //Start the remoting host
            ProcessStartInfo info = new ProcessStartInfo( "RemotingAdapterTests.ServerConsole.exe" );
            info.WindowStyle = ProcessWindowStyle.Hidden;
            _server = Process.Start( info );
        }

        [ClassCleanup( )]
        public static void MyClassCleanup( )
        {
            //Shutdown the remoting host
            if ( _server != null )
                _server.Kill( );

        }

        [TestMethod]
        [Owner( "Sergio Borromei" )]
        public void InvokeUsingRemoting( )
        {
            //Load remoting client configurations
            RemotingConfiguration.Configure( "client.config", false );

            //Get a remote instance of a service using a service contract
            ISalesServiceContract salesService = ( ISalesServiceContract )Activator.GetObject( typeof( ISalesServiceContract ), "tcp://localhost:8080/SalesService" );

            SalesOrderHeader order = salesService.GetSalesOrderById( 56343 );

            Assert.AreNotEqual( null, order );
            Assert.AreNotEqual( null, order.Customer );
            Assert.AreNotEqual( null, order.SalesOrderDetails );

            SalesOrderHeader order2 = salesService.GetSalesOrderById( 1111 );
            Assert.AreEqual( null, order2 );

            //Get a remote instance of a service without using a service contract
            SalesServiceWithoutContract salesServiceNoContract = ( SalesServiceWithoutContract )Activator.GetObject( typeof( SalesServiceWithoutContract ), "tcp://localhost:8080/SalesServiceWithoutContract" );

            SalesOrderHeader order3 = salesServiceNoContract.GetSalesOrderById( 126537 );

            Assert.AreNotEqual( null, order3 );
            Assert.AreNotEqual( null, order3.Customer );
            Assert.AreNotEqual( null, order3.SalesOrderDetails );

            SalesOrderHeader order4 = salesServiceNoContract.GetSalesOrderById( 1111 );
            Assert.AreEqual( null, order4 );
        }

        [TestMethod]
        [Owner( "Sergio Borromei" )]
        public void InitAdapterWithBadConfig( )
        {
            //Create a new RemotingAdapter instance
            IServiceAdapter adapter = new RemotingAdapter( );

            XmlDocument xml = new XmlDocument( );
            xml.LoadXml( "<badConfiguration/>" );

            try
            {
                adapter.Init( "", xml.DocumentElement );
            }
            catch ( Exception e )
            {
                Assert.AreEqual( typeof( InvocationFrameworkException ), e.GetType( ) );
            }

            try
            {
                adapter.Init( "SalesRemotingAdapter", xml.DocumentElement );
            }
            catch ( Exception e )
            {
                Assert.AreEqual( typeof( InvocationFrameworkException ), e.GetType( ) );
            }
        }

        [TestMethod]
        [Owner( "Sergio Borromei" )]
        public void InitAdapter( )
        {
            //Create a new RemotingAdapter instance
            IServiceAdapter adapter = new RemotingAdapter( );

            //Load remoting configuration
            XmlDocument xml = new XmlDocument( );
            xml.Load( "client.config" );

            //Initialize adapter with remoting config.
            adapter.Init( "SalesRemotingAdapter", xml.SelectSingleNode( "/configuration/system.runtime.remoting" ) );

            Assert.AreEqual( true, adapter.Initialized );
            Assert.AreEqual( "SalesRemotingAdapter", adapter.Id );
        }

        [TestMethod]
        [Owner( "Sergio Borromei" )]
        public void IsServiceAvailable( )
        {
            //Create a new RemotingAdapter instance
            IServiceAdapter adapter = new RemotingAdapter( );

            //Load remoting configuration
            XmlDocument xml = new XmlDocument( );
            xml.Load( "client.config" );

            //Initialize adapter with remoting config.
            adapter.Init( "SalesRemotingAdapter", xml.SelectSingleNode( "/configuration/system.runtime.remoting" ) );

            Assert.AreEqual( true, adapter.IsServiceAvailable( "SalesService" ) );
            Assert.AreEqual( true, adapter.IsServiceAvailable( "SalesServiceWithoutContract" ) );
        }

        [TestMethod]
        [Owner( "Sergio Borromei" )]
        public void IsServiceOperationAvailable( )
        {
            //Create a new RemotingAdapter instance
            IServiceAdapter adapter = new RemotingAdapter( );

            //Load remoting configuration
            XmlDocument xml = new XmlDocument( );
            xml.Load( "client.config" );

            //Initialize adapter with remoting config.
            adapter.Init( "SalesRemotingAdapter", xml.SelectSingleNode( "/configuration/system.runtime.remoting" ) );

            Assert.AreEqual( true, adapter.IsServiceOperationAvailable( "SalesService", "GetSalesOrderById" ) );
            Assert.AreEqual( true, adapter.IsServiceOperationAvailable( "SalesServiceWithoutContract", "GetSalesOrderById" ) );
            Assert.AreEqual( false, adapter.IsServiceOperationAvailable( "Nada", "Nada" ) );
            Assert.AreEqual( false, adapter.IsServiceOperationAvailable( "SalesService", "Nada" ) );

        }

        [TestMethod]
        [Owner( "Sergio Borromei" )]
        public void Execute( )
        {
            //Create a new RemotingAdapter instance
            IServiceAdapter adapter = new RemotingAdapter( );

            //Load remoting configuration
            XmlDocument xml = new XmlDocument( );
            xml.Load( "client.config" );

            //Initialize adapter with remoting config.
            adapter.Init( "SalesRemotingAdapter", xml.SelectSingleNode( "/configuration/system.runtime.remoting" ) );

            //Invoke service using contract
            object[] parameters = new object[ 1 ];
            parameters[ 0 ] = 56343;

            object result = adapter.Execute( "SalesService", "GetSalesOrderById", null, parameters );

            Assert.AreEqual( typeof( SalesOrderHeader ), result.GetType( ) );

            SalesOrderHeader order = ( SalesOrderHeader )result;
            Assert.AreNotEqual( null, order );
            Assert.AreNotEqual( null, order.Customer );
            Assert.AreNotEqual( null, order.SalesOrderDetails );


            //Invoke service without contract
            parameters[ 0 ] = 126537;
            object result2 = adapter.Execute( "SalesServiceWithoutContract", "GetSalesOrderById", null, parameters );
            Assert.AreEqual( typeof( SalesOrderHeader ), result2.GetType( ) );

            SalesOrderHeader order2 = ( SalesOrderHeader )result2;
            Assert.AreNotEqual( null, order2 );
            Assert.AreNotEqual( null, order2.Customer );
            Assert.AreNotEqual( null, order2.SalesOrderDetails );


            //Generate a server exception
            parameters[ 0 ] = 666;
            try
            {
                object result3 = adapter.Execute( "SalesService", "GetSalesOrderById", null, parameters );
            }
            catch ( Exception e )
            {
                Assert.AreEqual( typeof( DiabolicException ), e.GetType( ) );
            }
        }

        [TestMethod]
        [Owner( "Sergio Borromei" )]
        public void ExecuteUsingProxyFactory( )
        {
            //Invoke a service using a contract
            ServiceProxyBuilder<ISalesServiceContract> proxyBuilder = new ServiceProxyBuilder<ISalesServiceContract>( "ISalesServiceContract" );
            ISalesServiceContract salesService = proxyBuilder.BuildProxy( );

            try
            {
                SalesOrderHeader order = salesService.GetSalesOrderById( 56343 );
            }
            catch ( Exception e )
            {
                Assert.AreEqual( typeof( InvocationFrameworkInvocationException ), e.GetType( ) );
            }


            ServiceProxyBuilder<ISalesServiceContract> proxyBuilder1 = new ServiceProxyBuilder<ISalesServiceContract>( "SalesService" );
            salesService = proxyBuilder1.BuildProxy( );
            SalesOrderHeader order1 = salesService.GetSalesOrderById( 56343 );

            Assert.AreNotEqual( null, order1 );
            Assert.AreNotEqual( null, order1.Customer );
            Assert.AreNotEqual( null, order1.SalesOrderDetails );

            //Invoke a service without using a contract
            ServiceProxyBuilder<SalesServiceWithoutContract> proxyBuilder2 = new ServiceProxyBuilder<SalesServiceWithoutContract>( "SalesServiceWithoutContract" );
            SalesServiceWithoutContract salesServiceNoContract = proxyBuilder2.BuildProxy( );

            SalesOrderHeader order2 = salesServiceNoContract.GetSalesOrderById( 126537 );
            Assert.AreNotEqual( null, order2 );
            Assert.AreNotEqual( null, order2.Customer );
            Assert.AreNotEqual( null, order2.SalesOrderDetails );

            //Generate a server exception
            try
            {
                SalesOrderHeader order3 = salesService.GetSalesOrderById( 666 );
            }
            catch ( Exception e )
            {
                Assert.AreEqual( typeof( DiabolicException ), e.GetType( ) );
            }

            //Invoke the DataSet method
            SalesOrderDataSet orderDS = salesService.GetSalesOrderDataSet( 1 );

            Assert.IsNotNull( orderDS );
        }
   }
}